# core/dispatch.py
#
#

""" basic package for the program. """

__copyright__ = "Copyright 2014 B.H.J Thate"

## IMPORTS

from core.utils import get_plugname, get_name, get_strace
from core.object import Object

import collections
import threading
import logging
import random
import queue
import time

## RUNNER

class Runner(threading.Thread):

    """ the working unit of CORED.  """

    _nr = 0

    def __init__(zelf, *args, **kwargs):
        threading.Thread.__init__(zelf, None, zelf._loop, "Runner-%s" % str(zelf._nr), args, kwargs)
        zelf.setDaemon(True)
        zelf._queue = queue.Queue()
        zelf._ready = threading.Event() 
        zelf._status = "starting"
        zelf._nr += 1

    def _loop(zelf):
        """ loop that executes the (function, object) pairs. """
        args, kwargs = zelf._queue.get()
        func = args[0]
        try: arguments = args[1:]
        except IndexError: arguments = []
        name = get_name(func)
        zelf.setName(name)
        logging.warn("! %s" % name)
        func(*arguments, **kwargs)
        zelf.ready()

    def put(zelf, *args, **kwargs):
        """ put arguments/kwargs to the Runner. """
        zelf._queue.put((args, kwargs))
        return zelf

    ## WAITERS

    def ready(zelf):
        """ signal to ready state. """
        logging.debug("> ready %s" % get_name(zelf))
        zelf._ready.set()

    def clear(zelf):
        """ clear the ready state. """
        logging.debug("> clear %s" % get_name(zelf))
        zelf._ready.clear()

    def wait(zelf, sec=180.0):
        """ wait for ready state. """
        logging.debug("> wait %s %s" % get_name(zelf), get_strace())
        try: zelf._ready.wait(sec)
        except: pass


## DISPATCH

class Dispatcher(Object):

    """ the dispatcher delegates the workload to the Runners. Runner gets instantiated when needed. """

    def register(zelf, *args, **kwargs):
        """ callback type with corresponding callback function. """
        o = Object()
        o.cmnd = args[0]
        o.func = args[1]
        o.plugname = get_plugname()
        if o.cmnd not in zelf: zelf[o.cmnd] = []
        zelf[o.cmnd].append(o)

    def denied(zelf, *args, **kwargs):
        for obj in zelf.objects("denied"):
            if args[0] == obj.get("denied", None): return True
        return False

    def allowed(zelf, *args, **kwargs):
        for obj in zelf.objects("allowed"):
            if args[0] == obj.get("allowed", None): return True
        return False
 
    def dispatch(zelf, *args, **kwargs):
        """ dispatch an event onto a Runner. """
        event = args[0]
        cmnd = event.get_cmnd()
        if cmnd: return zelf.execute(cmnd, event)
        return 
 
    def execute(zelf, *args, **kwargs):
        """ execute a command, event pair. """
        from core import kernel
        cmndlist = args[0]
        event = args[1]
        if "origin" in event:
            if zelf.allowed(event.origin):  logging.warn("! allow %s" % event.origin)
            elif zelf.denied(event.origin): logging.warn("! deny %s" % event.origin) ; return
        result = []
        for cmnd in cmndlist:
            logging.info("> %s" % str(cmnd))
            result.append(cmnd.func(event))
        return result

    def run_func(zelf, *args, **kwargs):
        """ execute command/event pair. """
        cmnd = args[0]
        event = args[1]
        if cmnd not in zelf: return
        for functor in zelf[cmnd]: functor.func(event)
        event.ready()
        return event

    def put(zelf, *args, **kwargs):
        """ put load to the Runner. """
        logging.info("> put %s" % get_name(*args))
        target = zelf.make_new()
        return target.put(*args, **kwargs)

    def make_new(zelf, *args, **kwargs):
        """ create a Runner. """
        runner = Runner(*args, **kwargs)
        runner.start()
        return runner

